rem Oracle Cost Based Optimizer Basics Demo 3
rem 
rem (C) 2012 Randolf Geist 
rem http://oracle-randolf.blogspot.com
rem
rem This demo shows that data clustering / scattering makes a 
rem amazing difference to the efficiency of the "Small Job" approach
rem
rem It basically picks up the bad execution plan from Demo 1
rem which will take literally hours to complete on a normal
rem spinning disk I/O subsystem (using a sufficiently small buffer cache)
rem
rem The interesting point is that the cost estimate for the long running
rem version is exactly as the same as the cost estimate for a copy of T2
rem that is ordered in the same way as T1
rem
rem The optimizer does *not* have a clue about inter-table clustering
rem It only knows about intra-table clustering via the index clustering factor
rem
rem But look at the performance difference at actual execution time:
rem The bad execution plan only takes a couple of seconds rather than hours!
rem
rem The demo was tested and presented on 11.2

set echo on linesize 200 trimspool on trimout on tab off pagesize 999

-- Good clustering example
-- Get rid of any statistics that get the cardinality right
-- We want the bad execution plan back from Demo 1
exec dbms_stats.drop_extended_stats(null, 't1', '(trunc(attr1))')

exec dbms_stats.drop_extended_stats(null, 't1', '(trunc(attr2))')

-- Restore previous statistics from Demo 1
exec dbms_stats.gather_table_stats(null, 't1', method_opt => 'for all columns size 1 for columns size 254 attr1, attr2', no_invalidate => false)

-- This is the original cost estimate
explain plan for
select
        count(t2.attr2)
from
        t1
      , t2
where
/*------------------*/
        trunc(t1.attr1) = 1
and     trunc(t1.attr2) = 1
/*------------------*/
and     t1.fk = t2.id
;

select * from table(dbms_xplan.display);

-- Re-create T2
drop table t2;

purge table t2;

-- This time using the natural order of T1, not ordered by T1.FK
create table t2
as
select
        fk as id
      , ceil(rownum / 10) as attr1
      , ceil(rownum / 10) as attr2
      , rpad('x', 100) as filler
from
        (select rownum as rn, t1.* from t1)
order by
        rn
;

-- Basic statistics
exec dbms_stats.gather_table_stats(null, 't2', method_opt => 'for all columns size 1')

-- And again the same index on T2.ID
create unique index t2_idx on t2 (id);

set echo on timing on time on

-- The cost estimate is the same as before
explain plan for
select
        count(t2.attr2)
from
        t1
      , t2
where
/*------------------*/
        trunc(t1.attr1) = 1
and     trunc(t1.attr2) = 1
/*------------------*/
and     t1.fk = t2.id
;

select * from table(dbms_xplan.display);

alter session set statistics_level = all;

-- But what a dramatic difference at runtime!
select
        count(t2.attr2)
from
        t1
      , t2
where
/*------------------*/
        trunc(t1.attr1) = 1
and     trunc(t1.attr2) = 1
/*------------------*/
and     t1.fk = t2.id
;

@xplan_extended_display_cursor

set echo on timing on time on pagesize 999
