rem Oracle Cost Based Optimizer Basics Demo 2
rem 
rem (C) 2012 Randolf Geist 
rem http://oracle-randolf.blogspot.com
rem
rem This demo shows that Join Selectivities can be a tricky thing
rem
rem What seems to be obvious to you is *not* necessarily obvious
rem to the optimizer. It does *not* look at the data
rem
rem And for Join Selectivities unfortunately there is no Dynamic Sampling
rem or Cardinality Feedback available, simply because Oracle would have
rem to look at all possible combinations to make a fair comparison
rem between the estimates. It cannot just look at one particular join order
rem because the remaining join orders would then still be based on the 
rem original statistics based estimates
rem
rem Getting Join Selectivities right can be a tricky thing!
rem
rem The demo was tested and presented on 11.2

set echo on linesize 200 pagesize 999 tab off trimspool on trimout on

drop table t;

purge table t;

drop table d;

purge table d;

-- The FK column is highly skewed
-- A lot of rows (910K out of 1M) point to the same PK = 1
create table t
as
select
        rownum as id
      , case
        when rownum <= 910000
        then 1
        else mod(rownum, 999) + 2
        end as fk
      , rpad('x', 100) as filler
from
        dual
connect by
        level <= 1000000
;

-- No histogram on FK yet!
exec dbms_stats.gather_table_stats(null, 't', method_opt => 'for all columns size 1')

-- Index that allows the "Small Job" approach to be evaluated
create index t_idx on t (fk);

-- The "Parent" table with just 1K rows
create table d
as
select
        rownum as id
      , 'DESC' || rownum as some_attr
      , rpad('x', 100) as filler
from
        dual
connect by
        level <= 1000
;

-- Simple table stats
exec dbms_stats.gather_table_stats(null, 'd', method_opt => 'for all columns size 1')

-- Proper PK/FK constraints
alter table d add constraint d_pk primary key (id);

alter table t add constraint t_fk_d foreign key (fk) references d(id);

-- Data distribution overview of T
select
        *
from
        (
          select
                  fk
                , count(*) as cnt
          from
                  t
          group by
                  fk
          order by
                  cnt desc
                , fk
        )
where
        rownum <= 10
;

-- Data distribution overview of D
select
        *
from
        (
          select
                  id
                , some_attr
          from
                  d
          order by
                  id
        )
where
        rownum <= 10
;

-- Basic column statistics of T.FK
select
        num_distinct
      , utl_raw.cast_to_number(low_value) as low_value
      , utl_raw.cast_to_number(high_value) as high_value
      , num_buckets
      , sample_size
      , histogram
      , last_analyzed
from
        user_tab_col_statistics
where
        table_name = 'T'
and     column_name = 'FK'
;

-- Although we can see that filtering on a single row 
-- D.SOME_ATTR = 'DESC1' and joining the corresponding
-- PK = 1 to T will match 910K rows in T, the optimizer
-- cannot tell this from the available statistics
-- It simply sees 1K distinct values in T.FK
-- So that makes on average 1K rows out of 1M rows in T
-- if I filter on 1 row out of 1K rows in D
explain plan for
select
       count(t.filler)
from
       t
     , d
where
       d.some_attr = 'DESC1'
and    t.fk = d.id
;

select * from table(dbms_xplan.display(format => 'BASIC PREDICATE ROWS'));

-- But wait, the histogram on T.FK is missing!
explain plan for
select
       count(*)
from
       t
where
       t.fk = 1
;

select * from table(dbms_xplan.display(format => 'BASIC PREDICATE ROWS'));

-- So let's add it
exec dbms_stats.gather_table_stats(null, 't', method_opt => 'for all columns size 1 for columns size 254 fk')

-- Column statistics for T.FK tell us there is a histogram now
select
        num_distinct
      , utl_raw.cast_to_number(low_value) as low_value
      , utl_raw.cast_to_number(high_value) as high_value
      , num_buckets
      , sample_size
      , histogram
      , last_analyzed
from
        user_tab_col_statistics
where
        table_name = 'T'
and     column_name = 'FK'
;

-- The histogram data
-- Clearly 1 is now a very popular value in T.FK to the optimizer
select
        cum_cnt - lag(cum_cnt, 1, 0) over (order by endpoint_value) as cnt
      , endpoint_value
from
        (
          select
                  round(h.endpoint_number/c.num_buckets*t.sample_size) as cum_cnt
                , endpoint_value
          from
                  user_tab_histograms h
                , user_tables t
                , user_tab_columns c
          where
                  h.table_name = 'T'
          and     h.column_name = 'FK'
          and     t.table_name = h.table_name
          and     c.table_name = h.table_name
          and     c.column_name = h.column_name
        )
;

-- Which is confirmed here again
explain plan for
select
       count(*)
from
       t
where
       t.fk = 1
;

select * from table(dbms_xplan.display(format => 'BASIC PREDICATE ROWS'));

-- Yikes! Still the same 1K rows bad cardinality estimate
-- Even in the presence of a histogram on T.FK
--
-- The optimizer does *not* know that D.SOME_ATTR = 'DESC1' maps to D.ID = 1
-- It cannot tell this from the statistics available
-- This renders the histogram useless in this case
explain plan for
select
       count(t.filler)
from
       t
     , d
where
       d.some_attr = 'DESC1'
and    t.fk = d.id
;

select * from table(dbms_xplan.display(format => 'BASIC PREDICATE ROWS'));

alter session set statistics_level = all;

-- We really get 910K rows back
-- For those that don't believe above said
select
       count(t.filler)
from
       t
     , d
where
       d.some_attr = 'DESC1'
and    t.fk = d.id
;

@xplan_extended_display_cursor

-- Some ideas how the cardinality estimate can be corrected
set doc off
doc
-- Emulating the SQL Tuning Advisor with undocumented OPT_ESTIMATE hints
explain plan for
select /*+ opt_estimate(table d rows=900) opt_estimate(join (t d) rows=910000) */
       count(t.filler)
from
       t
     , d
where
       d.some_attr = 'DESC1'
and    t.fk = d.id
;

select * from table(dbms_xplan.display);

-- Filtering on D.ID and Transitive Closure allows 
-- the optimizer to get it right
-- But we do not filter on D.ID in our original case
explain plan for
select
       count(t.filler)
from
       t
     , d
where
       d.id = 1
and    t.fk = d.id
;

select * from table(dbms_xplan.display);

-- Another undocumented feature
-- Let's Oracle evaluate the subquery at optimization time
-- and rewrite the query so that the optimizer knows about
-- the actual value of D.ID
explain plan for
select
       count(t.filler)
from
       t
where
       fk in
       (
         select /*+ precompute_subquery */
                id
         from
                d
         where
                d.some_attr = 'DESC1'
       )
;

select * from table(dbms_xplan.display);

-- Another approach
-- Fudging statistics on D.SOME_ATTR
-- But be aware of the side effects to other queries!
declare
 srec dbms_stats.statrec;
 novals dbms_stats.chararray;
 distcnt number;
 density number;
 nullcnt number;
 avgclen number;
begin
  sys.dbms_stats.get_column_stats(
  ownname=>user,
  tabname=>'D',
  colname=>'SOME_ATTR',
  distcnt => distcnt,
  density=>density,
  nullcnt=>nullcnt,
  srec=>srec,
  avgclen=>avgclen
  );
  srec.epc := 2;
  novals := dbms_stats.chararray(
  'DESC0',
  'DESC1'
  );
  srec.bkvals := dbms_stats.numarray(
  1,
  910
  );
  --
  dbms_stats.prepare_column_values (srec,novals);
  --
  dbms_stats.set_column_stats(
  ownname=>user,
  tabname=>'D',
  colname=>'SOME_ATTR',
  distcnt=>1000,
  nullcnt=>90,
  srec=>srec,
  avgclen=>avgclen
  );
end;
/

-- Now this plan looks better
-- But other plans might look worse now!
explain plan for
select
       count(t.filler)
from
       t
     , d
where
       d.some_attr = 'DESC1'
and    t.fk = d.id
;

select * from table(dbms_xplan.display);
#
