rem Oracle Cost Based Optimizer Basics Demo 1
rem 
rem (C) 2012 Randolf Geist 
rem http://oracle-randolf.blogspot.com
rem
rem This demo shows how a wrong Single Table Cardinality estimate
rem ruins a simple two table join
rem
rem Because Oracle estimates that the "Small Job" approach
rem using an index driven access is more efficient than the
rem "Big Job" approach. But this estimate is wrong because
rem the Single Table Cardinality estimate is way off (100 vs. 900K)
rem
rem Although it does *not* change the Single Table Access method 
rem for the table, since there is only a TABLE ACCESS FULL possible
rem
rem The moral of the story:
rem Make sure that the Single Table Cardinality estimates are in the 
rem right ballpark. It is easy to check and verify and there are
rem various features available to support the optimizer (Dynamic Sampling,
rem Extended Statistics, Virtual Columns, Function-Based Indexes 
rem to name a few)
rem
rem In order to see the dramatic effect of the bad clustering of data
rem between T1 and T2 with the "Small Job" approach
rem you need a small buffer cache, like 64MB
rem Since the tables are both pretty small (128MB)
rem
rem The demo was tested and presented on 11.2

set echo on linesize 200 trimspool on trimout on tab off pagesize 999

drop table t1;

purge table t1;

drop table t2;

purge table t2;

-- Initialize the random generator for "reproducible" pseudo-randomness
exec dbms_random.seed(0)

-- Create the first table
-- Note the FK column pointing to T2 is basically randomized
create table t1
as
select
        rownum as id
      , trunc(dbms_random.value(1, 1000000000000)) as fk
      , case when rownum <= 900000 then 1 else ceil(rownum / 10) end as attr1
      , case when rownum <= 900000 then 1 else ceil(rownum / 10) end as attr2
      , rpad('x', 100) as filler
from
        dual
connect by
        level <= 1000000
;

-- Make the optimizer aware of the highly skewed data in ATTR1 and ATTR2
-- Note that in 11g you could also create a column group (ATTR1, ATTR2)
-- to make the optimizer aware of the 100% correlated data in those columns
exec dbms_stats.gather_table_stats(null, 't1', method_opt => 'for all columns size 1 for columns size 254 attr1, attr2')

-- T2 uses the FK from T1 as "primary key"
-- But is ordered by ID, so different order from T1.FK
create table t2
as
select
        fk as id
      , ceil(rownum / 10) as attr1
      , ceil(rownum / 10) as attr2
      , rpad('x', 100) as filler
from
        t1
order by
        id
;

-- Simple basic column and table statistics
exec dbms_stats.gather_table_stats(null, 't2', method_opt => 'for all columns size 1')

-- The index that enables the evaluation of the "Small Job" approach
create unique index t2_idx on t2 (id);

set echo on timing on time on

-- A look at the data distribution of ATTR1 and ATTR2 in T1
select
        *
from
        (
          select
                  attr1
                , attr2
                , count(*) as cnt
          from
                  t1
          group by
                  attr1
                , attr2
          order by
                  cnt desc
                , attr1
                , attr2
        )
where
        rownum <= 10
;

-- Enable Rowsource statistics with proper time measurement (overhead!)
alter session set statistics_level = all;

-- A sample execution of the "Big Job" approach
-- Notice how the optimizer gets the cardinality of T1 in the right ballpark
select
        count(t2.attr2)
from
        t1
      , t2
where
/*------------------*/
        t1.attr1 = 1
and     t1.attr2 = 1
/*------------------*/
and     t1.fk = t2.id
;

@xplan_extended_display_cursor

set echo on timing on time on pagesize 999

-- A sample execution of the "Small Job" approach
-- Notice how the optimizer gets the cardinality of T1 in the right ballpark
select
        count(t2.attr2)
from
        t1
      , t2
where
/*------------------*/
        t1.attr1 = 90001
and     t1.attr2 = 90001
/*------------------*/
and     t1.fk = t2.id
;

@xplan_extended_display_cursor

set echo on timing on time on pagesize 999

-- Compare the ACCESS path for the original filter predicates
explain plan for
select
        count(*)
from
        t1
where
        attr1 = 1
and     attr2 = 1;

select * from table(dbms_xplan.display(format => 'BASIC PREDICATE ROWS'));

-- to the new ones using the TRUNC() function
-- Notice how the TABLE ACCESS FULL remains the same
-- But look at the estimated ROWS
-- Oracle loses track of the column data skew due to the functions
explain plan for
select
        count(*)
from
        t1
where
        trunc(attr1) = 1
and     trunc(attr2) = 1;

select * from table(dbms_xplan.display(format => 'BASIC PREDICATE ROWS'));

-- The actual data distribution in ATTR1 and ATTR2 remains the same
-- Which is obvious to us, but not to the optimizer
-- It doesn't look at the actual data by default
-- If it can't tell from the statistics it starts to guess
select
        *
from
        (
          select
                  trunc(attr1)
                , trunc(attr2)
                , count(*) as cnt
          from
                  t1
          group by
                  trunc(attr1)
                , trunc(attr2)
          order by
                  cnt desc
                , trunc(attr1)
                , trunc(attr2)
        )
where
        rownum <= 10
;

-- Things go very wrong here with a small buffer cache
-- T2 rows are badly scattered when accessed via T1.FK
--
-- The "Small Job" approach would have been a good idea
-- if the estimate of 100 loop iterarions was correct!
select
        count(t2.attr2)
from
        t1
      , t2
where
/*------------------*/
        trunc(t1.attr1) = 1
and     trunc(t1.attr2) = 1
/*------------------*/
and     t1.fk = t2.id
;

@xplan_extended_display_cursor "" "" "ALLSTATS LAST +COST"

set echo on timing on time on pagesize 999

-- Some ideas how things could be corrected
set doc off
doc
exec dbms_stats.gather_table_stats(null, 't1', method_opt => 'for all columns size 1 for columns size 254 attr1, attr2, (trunc(attr1)), (trunc(attr2))', no_invalidate => false)

select
        count(t2.attr2)
from
        t1
      , t2
where
/*------------------*/
        trunc(t1.attr1) = 1
and     trunc(t1.attr2) = 1
/*------------------*/
and     t1.fk = t2.id
;

@xplan_extended_display_cursor

set echo on timing on time on pagesize 999

select  /*+ dynamic_sampling(t1 7) */
        count(t2.attr2)
from
        t1
      , t2
where
/*------------------*/
        trunc(t1.attr1) = 1
and     trunc(t1.attr2) = 1
/*------------------*/
and     t1.fk = t2.id
;

@xplan_extended_display_cursor

set echo on timing on time on pagesize 999
#
